@@ -16,6 +16,7 @@
 #include "terminal_ui.hh"
 #include "option_types.hh"
 #include "parameters_parser.hh"
+#include "profile.hh"
 #include "ranges.hh"
 #include "regex.hh"
 #include "register_manager.hh"
@@ -44,6 +45,19 @@ struct {
     unsigned int version;
     StringView notes;
 } constexpr version_notes[] = { {
+        20240509,
+        "» {+u}flag-lines -after{} highlighter\n"
+        "» asynchronous {+u}shell-script-candidates{} completion\n"
+        "» {+b}%val\\{window_range}{} is now emitted as separate strings\n"
+        "» {+b}+{} only duplicates identical selections a single time\n"
+        "» {+u}daemonize-session{} command\n"
+        "» view mode and mouse scrolling no longer change selections\n"
+        "» {+u}git apply/blame-jump/edit/grep{} commands\n"
+        "» {+u}git blame{} works in {+u}git-diff{} and {+u}git-log{} buffers\n"
+        "» custom completions are no longer sorted if the typed text is empty\n"
+        "» {+u}terminal{} now selects implementation based on windowing options\n"
+        "» {+u}local{} scopes\n"
+    }, {
         20230805,
         "» Fix FreeBSD/MacOS clang compilation\n"
     }, {
@@ -221,7 +235,7 @@ static const EnvVarDesc builtin_env_vars[] = { {
     }, {
         "buflist", false,
         [](StringView name, const Context& context) -> Vector<String>
-        { return BufferManager::instance() | transform(&Buffer::display_name) | gather<Vector>(); }
+        { return BufferManager::instance() | transform(&Buffer::display_name) | gather<Vector<String>>(); }
     }, {
         "buf_line_count", false,
         [](StringView name, const Context& context) -> Vector<String>
@@ -289,7 +303,7 @@ static const EnvVarDesc builtin_env_vars[] = { {
         [](StringView name, const Context& context) -> Vector<String>
         { return ClientManager::instance() |
                       transform([](const std::unique_ptr<Client>& c) -> const String&
-                                { return c->context().name(); }) | gather<Vector>(); }
+                                { return c->context().name(); }) | gather<Vector<String>>(); }
     }, {
         "modified", false,
         [](StringView name, const Context& context) -> Vector<String>
@@ -335,21 +349,21 @@ static const EnvVarDesc builtin_env_vars[] = { {
         { return main_sel_first(context.selections()) |
                      transform([&buffer=context.buffer()](const Selection& sel) {
                          return selection_to_string(ColumnType::Byte, buffer, sel);
-                     }) | gather<Vector>(); }
+                     }) | gather<Vector<String>>(); }
     }, {
         "selections_char_desc", false,
         [](StringView name, const Context& context) -> Vector<String>
         { return main_sel_first(context.selections()) |
                      transform([&buffer=context.buffer()](const Selection& sel) {
                          return selection_to_string(ColumnType::Codepoint, buffer, sel);
-                     }) | gather<Vector>(); }
+                     }) | gather<Vector<String>>(); }
     }, {
         "selections_display_column_desc", false,
         [](StringView name, const Context& context) -> Vector<String>
         { return main_sel_first(context.selections()) |
                      transform([&buffer=context.buffer(), tabstop=context.options()["tabstop"].get<int>()](const Selection& sel) {
                          return selection_to_string(ColumnType::DisplayColumn, buffer, sel, tabstop);
-                     }) | gather<Vector>(); }
+                     }) | gather<Vector<String>>(); }
     }, {
         "selection_length", false,
         [](StringView name, const Context& context) -> Vector<String>
@@ -381,9 +395,9 @@ static const EnvVarDesc builtin_env_vars[] = { {
         "window_range", false,
         [](StringView name, const Context& context) -> Vector<String>
         {
-            auto setup = context.window().compute_display_setup(context);
-            return {format("{} {} {} {}", setup.first_line, setup.first_column,
-                                          setup.line_count, 0)};
+            const auto& setup = context.window().last_display_setup();
+            return {to_string(setup.first_line), to_string(setup.first_column),
+                    to_string(setup.line_count), to_string(0)};
         }
     }, {
         "history", false,
@@ -599,7 +613,6 @@ void register_options()
 }
 
 static Client* local_client = nullptr;
-static int local_client_exit = 0;
 static bool convert_to_client_pending = false;
 
 enum class UIType
@@ -667,38 +680,7 @@ pid_t fork_server_to_background()
 
 std::unique_ptr<UserInterface> create_local_ui(UIType ui_type)
 {
-    if (ui_type != UIType::Terminal)
-        return make_ui(ui_type);
-
-    struct LocalUI : TerminalUI
-    {
-        LocalUI()
-        {
-            set_signal_handler(SIGTSTP, [](int) {
-                if (ClientManager::instance().count() == 1 and
-                    *ClientManager::instance().begin() == local_client)
-                    TerminalUI::instance().suspend();
-                else
-                    convert_to_client_pending = true;
-           });
-        }
-
-        ~LocalUI() override
-        {
-            local_client = nullptr;
-            if (convert_to_client_pending or
-                ClientManager::instance().empty())
-                return;
-
-            if (fork_server_to_background())
-            {
-                this->TerminalUI::~TerminalUI();
-                exit(local_client_exit);
-            }
-        }
-    };
-
-    if (not isatty(0))
+    if (ui_type == UIType::Terminal and not isatty(0))
     {
         // move stdin to another fd, and restore tty as stdin
         int fd = dup(0);
@@ -708,7 +690,19 @@ std::unique_ptr<UserInterface> create_local_ui(UIType ui_type)
         create_fifo_buffer("*stdin*", fd, Buffer::Flags::None);
     }
 
-    return std::make_unique<LocalUI>();
+    auto ui = make_ui(ui_type);
+
+    static SignalHandler old_handler = set_signal_handler(SIGTSTP, [](int sig) {
+        if (ClientManager::instance().count() == 1 and
+            *ClientManager::instance().begin() == local_client)
+            old_handler(sig);
+        else
+        {
+            convert_to_client_pending = true;
+            set_signal_handler(SIGTSTP, old_handler);
+        }
+    });
+    return ui;
 }
 
 int run_client(StringView session, StringView name, StringView client_init,
@@ -806,14 +800,11 @@ int run_server(StringView session, StringView server_init,
     write_to_debug_buffer("*** This is the debug buffer, where debug info will be written ***");
 
 #ifdef KAK_DEBUG
-    const auto start_time = Clock::now();
-    UnitTest::run_all_tests();
-
-    if (debug_flags & DebugFlags::Profile)
     {
-        using namespace std::chrono;
-        write_to_debug_buffer(format("running the unit tests took {} ms",
-                                     duration_cast<milliseconds>(Clock::now() - start_time).count()));
+        ProfileScope profile{debug_flags, [&](std::chrono::microseconds duration) {
+            write_to_debug_buffer(format("running the unit tests took {} ms", duration.count()));
+        }};
+        UnitTest::run_all_tests();
     }
 #endif
 
@@ -874,13 +865,14 @@ int run_server(StringView session, StringView server_init,
          write_to_debug_buffer(format("error while opening command line files: {}", error.what()));
     }
 
+    int exit_status = 0;
     try
     {
         if (not server.is_daemon())
         {
             local_client = client_manager.create_client(
                  create_local_ui(ui_type), getpid(), {}, get_env_vars(), client_init, init_buffer, std::move(init_coord),
-                 [](int status) { local_client_exit = status; });
+                 [&](int status) { exit_status = status; });
 
             if (startup_error and local_client)
                 local_client->print_status({
@@ -919,24 +911,22 @@ int run_server(StringView session, StringView server_init,
             global_scope.option_registry().clear_option_trash();
 
             if (local_client and not contains(client_manager, local_client))
-                local_client = nullptr;
-            else if (local_client and not local_client->is_ui_ok())
             {
-                ClientManager::instance().remove_client(*local_client, false, -1);
                 local_client = nullptr;
-                if (not client_manager.empty() and fork_server_to_background())
-                    return 0;
+                if ((not client_manager.empty() or server.is_daemon()) and fork_server_to_background())
+                    exit(exit_status); // We do not want to run destructors and hooks here
             }
             else if (convert_to_client_pending)
             {
                 kak_assert(local_client);
                 auto& local_context = local_client->context();
-                const String client_name = local_context.name();
-                const String buffer_name = local_context.buffer().name();
-                const String selections = selection_list_to_string(ColumnType::Byte, local_context.selections());
+                String client_name = local_context.name();
+                String buffer_name = local_context.buffer().name();
+                String selections = selection_list_to_string(ColumnType::Byte, local_context.selections());
 
                 ClientManager::instance().remove_client(*local_client, true, 0);
                 client_manager.clear_client_trash();
+                local_client = nullptr;
                 convert_to_client_pending = false;
 
                 if (fork_server_to_background())
@@ -951,7 +941,7 @@ int run_server(StringView session, StringView server_init,
     }
     catch (const kill_session& kill)
     {
-        local_client_exit = kill.exit_status;
+        exit_status = kill.exit_status;
     }
 
     {
@@ -959,7 +949,7 @@ int run_server(StringView session, StringView server_init,
         global_scope.hooks().run_hook(Hook::KakEnd, "", empty_context);
     }
 
-    return local_client_exit;
+    return exit_status;
 }
 
 int run_filter(StringView keystr, ConstArrayView<StringView> files, bool quiet, StringView suffix_backup)
@@ -1228,7 +1218,7 @@ int main(int argc, char* argv[])
             }
             String new_files;
             for (auto name : files) {
-                new_files += format("edit '{}'", escape(real_path(name), "'", '\\'));
+                new_files += format("edit '{}'", escape(real_path(name), "'", '\''));
                 if (init_coord) {
                     new_files += format(" {} {}", init_coord->line + 1, init_coord->column + 1);
                     init_coord.reset();
